import {
    Vector,
    world, system,
    removeEntity,
    clamp,
    PI_180,
    ItemStack
} from 'mhafy1016.lib/Util';

Number.prototype.inRange = function (min, max) {
    return (this >= min && this <= max);
}

const MAX_SPEED = 0.08, SPEEDRATE = 0.005;

function tick() {
    for (let player of world.getPlayers()) {
        try {
            if (player.scriptRetry && player.scriptRetry >= 3) continue;

            if (player.getEquipment('Chest') && player.getEquipment('Chest').typeId.startsWith('ewings:')) {
                if (!player.fmngr) {
                    if (player.getDynamicProperty('ewings:fmngr_id')) {
                        let fmngr = world.getEntity(player.getDynamicProperty('ewings:fmngr_id'));
                        if (!(fmngr && fmngr.isValid())) {
                            player.setDynamicProperty('ewings:fmngr_id', null);
                            continue;
                        }
                        player.fmngr = fmngr;
                    }

                    else player.fmngr = player.dimension.spawnEntity('ewings:fmngr', player.location);

                    player.setDynamicProperty('ewings:fmngr_id', player.fmngr.id);
                    player.fmngr.addTag(`owner=${player.id}`);

                } else {

                    if (!player.fmngr.isValid()) {
                        player.setDynamicProperty('ewings:fmngr_id', null);
                        delete player.fmngr;
                        continue;
                    }

                    let speed = player.fmngr.getFlyingSpeed();

                    if (!player.hasTag('isFlying')) {
                        if (player.isJumping) {
                            if (!player.hasJumped && system.currentTick - player.jumpTick < 5) {
                                player.runCommandAsync(`ride @s start_riding @e[type=ewings:fmngr,tag="owner=${player.id}"] teleport_rider`);
                            }
                            player.jumpTick = system.currentTick;
                            player.hasJumped = true;
                        } else player.hasJumped = false;
                    }

                    let ride = player.getRide();
                    if (ride && ride.isValid() && ride.typeId === 'ewings:fmngr') {
                        player.triggerEvent('ewings:gravity_off');
                        player.addTag('isFlying');
                        
                        const r = { x: -player.getRotation().x * PI_180, y: -player.getRotation().y * PI_180 };
                        const target = ride.v3().add(MAX_SPEED * (Math.sin(r.y) * Math.cos(r.x)), MAX_SPEED * Math.sin(r.x), MAX_SPEED * (Math.cos(r.y) * Math.cos(r.x)));
                        const target_v = ride.v3().getDirection(target);

                        speed = clamp(speed + SPEEDRATE, -MAX_SPEED, MAX_SPEED);
                        speed -= clamp(speed * calc(ride.getVelocity(), target_v) * 0.2, 0, MAX_SPEED);
                        const diff = ride.getVelocity().y - target_v.y;
                        const ny = Math.max(1, ride.getVelocity().y) * speed;
                        const velocity = Vector.multiply(target_v, speed);

                        if (Math.abs(diff) > Math.abs(target_v.y)) velocity.y = (-diff) * (ny * 3);
                        ride.applyImpulse(velocity);
                        ride.setRotation({ x: 0, y: player.getRotation().y });

                        try {
                            let below = ride.dimension.getBlock(ride.v3().sub(0, 0.05, 0));
                            if (below && below.isValid()) {
                                if (ride.getVelocity().y.inRange(-0.01, 0) && below.typeId != 'minecraft:air') {
                                    ride.runCommandAsync(`ride @s evict_riders`);
                                    player.removeTag('isFlying');
                                    player.triggerEvent('ewings:gravity_on');
                                }
                            }
                        } catch (err) { }

                        if (!player.playFlyingSound && speed >= MAX_SPEED) {
                            player.playSound(`wings.flying`);
                            player.playFlyingSound = true;
                        }
                        if (system.currentTick % 800 == 0) delete player.playFlyingSound;
                    } else {
                        speed = 0.04;
                        if (player.hasTag('isFlying')) {

                            if (player.isSneaking || player.isJumping) {
                                player.triggerEvent('ewings:gravity_on');
                                if (player.isSneaking)
                                    player.addEffect('slow_falling', 20, { amplifier: 0, showParticles: false });

                                if (player.isJumping) {
                                    player.addEffect('levitation', 20, { amplifier: 5, showParticles: false });
                                    if (!player.hasJumped && system.currentTick - player.jumpTick < 5) {
                                        player.runCommandAsync(`ride @s start_riding @e[type=ewings:fmngr,tag="owner=${player.id}"] teleport_rider`);
                                        player.hasJumped = false;
                                        continue;
                                    }
                                    player.jumpTick = system.currentTick;
                                    player.hasJumped = true;
                                }
                            } else {
                                player.hasJumped = false;
                                player.triggerEvent('ewings:gravity_off');
                                player.removeEffect('slow_falling');
                                player.removeEffect('levitation');

                                if (player.location.y != player._y) player.teleport(player.location);
                                player._y = player.location.y;
                            }

                            if (player.isOnGround && player.getVelocity().y == 0) {
                                player.removeTag('isFlying');
                                player.triggerEvent('ewings:gravity_on');
                            }
                        }

                        player.fmngr.teleport(player.v3(), { rotation: player.getRotation() });
                        if (player.playFlyingSound) player.runCommandAsync(`stopsound @s`);
                        delete player.playFlyingSound;
                    }

                    player.fmngr.setFlyingSpeed(speed);
                }
            } else {
                if (player.fmngr && player.fmngr.isValid()) {
                    player.fmngr.remove();
                    player.setDynamicProperty('ewings:fmngr_id', null);
                    delete player.fmngr;
                } else if (player.getDynamicProperty('ewings:fmngr_id')) {
                    removeEntity(player.getDynamicProperty('ewings:fmngr_id'));
                }
                player.triggerEvent('ewings:gravity_on');
                player.removeTag('isFlying');
            }

            //player.onScreenDisplay.setActionBar(`${str}`);
        } catch (err) {
            player.scriptRetry = (player.scriptRetry || 0) + 1;
            player.runCommandAsync(`event entity @e[type=ewings:fmngr,tag="owner=${player.id}"] ewings:despawn`);
            player.setDynamicProperty('ewings:fmngr_id', null);
            player.triggerEvent('ewings:gravity_on');
            player.removeEffect('slow_falling');
            player.removeEffect('levitation');
            player.removeTag('isFlying');
            delete player.fmngr;
            delete player.hasJumped;
            delete player.jumpTick;

            throw `${player.nameTag} has dealt an error\n ${3 - player.scriptRetry} tries left \n ${err}`;
        }
    }
}

system.runInterval(tick, 0);

function calc(a, b) {
    let a_mag = Math.sqrt(Math.pow(a.x, 2) + Math.pow(a.z, 2));
    let b_mag = Math.sqrt(Math.pow(b.x, 2) + Math.pow(b.z, 2));
    return Math.abs(~~(Math.acos((a.x * b.x + a.z * b.z) / (a_mag * b_mag)) * 180 / Math.PI) / 180);
}

world.afterEvents.entityDie.subscribe(e => {
    let n = Math.random(), bonus_chance = 0;
    let attackerId = '', projectileId = '', cause = '';

    if (e.damageSource.damagingEntity) attackerId = e.damageSource.damagingEntity.typeId;
    if (e.damageSource.damagingProjectile) projectileId = e.damageSource.damagingProjectile.typeId;
    if (e.damageSource.cause) cause = e.damageSource.cause;

    if (attackerId == 'minecraft:player') bonus_chance = 0.05;
    else if (attackerId == 'minecraft:creeper' && cause == 'entityExplosion') bonus_chance = 0.1;
    else if (attackerId == 'minecraft:skeleton' && cause == 'projectile') bonus_chance = 0.15;
    else if (attackerId == 'minecraft:zombie' && projectileId == 'minecraft:trident') bonus_chance = 0.2;

    let id = `${e.deadEntity.typeId}`;

    if (id.includes('bat') && n < 0.1 + bonus_chance)// 10% drop rate + bonus chance
        e.deadEntity.dimension.spawnItem(new ItemStack('ewings:bat_0'), e.deadEntity.location);
    else if (id.includes('vex') && n < 0.02 + bonus_chance)// 2% drop rate + bonus chance
        e.deadEntity.dimension.spawnItem(new ItemStack('ewings:butterfly_0'), e.deadEntity.location);
    else if (id.includes('bee') && n < 0.05 + bonus_chance)// 5% drop rate + bonus chance
        e.deadEntity.dimension.spawnItem(new ItemStack('ewings:dragonfly_0'), e.deadEntity.location);
    else if (id.includes('chicken') && n < bonus_chance)// 0% drop rate + bonus chance
        e.deadEntity.dimension.spawnItem(new ItemStack('ewings:feather_0'), e.deadEntity.location);
    else if ((id.includes('blaze') || id.includes('ghast') || id.includes('wither')) && n < bonus_chance)// 0% drop rate + bonus chance
        e.deadEntity.dimension.spawnItem(new ItemStack('ewings:ender_0'), e.deadEntity.location);
});