// Created by Mhafy1016
// Mhafy.1016#5886

import * as mc from '@minecraft/server';

export const Entity = mc.Entity;
export const Block = mc.Block;
export const BlockPermutation = mc.BlockPermutation;
export const ItemStack = mc.ItemStack;
export const world = mc.world;
export const system = mc.system;
export const Vector = mc.Vector;

export const PI_180 = Math.PI / 180;

const ALPHABET = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ', NUMBERS = '0123456789', SPECIAL = '!@#$%^&*()_-+=<>?';

export const sleep = (delay) => new Promise((resolve) => system.runTimeout(resolve, delay));

Entity.prototype.addItem = function (item) {
    if (this.hasComponent('inventory')) return this.getComponent('inventory').container.addItem(item);
}

Entity.prototype.clearEffects = function () {
    for (let e of this.getEffects()) {
        this.removeEffect(e.typeId);
    }
    return true;
}

Entity.prototype.setItem = function (item, slot) {
    if (!this.hasComponent('inventory')) return;
    let itemStack = item;
    if (typeof item === 'string') {
        if (item === 'minecraft:air' || item === 'air' || item === '') return this.clearItem(slot);
        else if (!item.includes(':')) itemStack = new ItemStack(`minecraft:` + item);
        else itemStack = new ItemStack(item);
    }
    let s = this.selectedSlot;
    if (!isNaN(slot)) s = slot;
    this.getComponent('inventory').container.setItem(s, itemStack);
}

Entity.prototype.clearItem = function (slot) {
    let s = slot || this.selectedSlot || 0;
    let slot_type = 'slot.hotbar';
    if (s >= 9) {
        slot_type = 'slot.inventory';
        s -= 9;
    }
    this.runCommandAsync(`replaceitem entity @s ${slot_type} ${s} air`);
}

Entity.prototype.getItem = function (slot) {
    let s = slot || this.selectedSlot || 0;
    if (this.hasComponent('inventory')) return this.getComponent('inventory').container.getItem(s);
    return null;
}

Entity.prototype.findItem = function (id) {
    let list = [];
    let con = this.getComponent('inventory').container;
    for (let i = 0; i < con.size; i++) {
        if (con.getItem(i) && con.getItem(i).typeId.includes(id)) {
            list.push({
                item: con.getItem(i),
                slot: i
            });
        }
    }
    return list;
}

Entity.prototype.getEquipment = function (equipmentSlot) {
    if (this.hasComponent("minecraft:equippable"))
        return this.getComponent("minecraft:equippable").getEquipment(equipmentSlot);
    return;
}

Entity.prototype.isSheared = function () {
    return this.getComponent("is_sheared") || null;
}

Entity.prototype.isIgnited = function () {
    return this.getComponent("is_ignited") || null;
}

Entity.prototype.isItem = function () {
    return this.getComponent("item") || null;
}

Entity.prototype.markVariant = function () {
    return this.getComponent('mark_variant') || null;
}
Entity.prototype.variant = function () {
    return this.getComponent('variant') || null;
}

Entity.prototype.skin_id = function () {
    return this.getComponent('skin_id') || null;
}

Entity.prototype.scale = function () {
    return this.getComponent('scale') || null;
}

Entity.prototype.getRide = function () {
    try {
        if (!this.hasComponent('riding') || !this.getComponent('riding')) return null;
        return this.getComponent('riding').entityRidingOn;
    } catch (err) { }
    return null;
}

Entity.prototype.isRiding = function (id) {
    try {
        if (!this.hasComponent('riding') || !this.getComponent('riding')) return null;
        let ride = this.getComponent('riding').entityRidingOn;
        if (ride && ride.isValid()) {
            if (typeof id === 'string' && ride.typeId === id) return ride;
            if (!isNaN(id) && ride.id == id) return ride;
        }
    } catch (err) { }
    return null;
}

Entity.prototype.isRideable = function () {
    return this.getComponent('rideable') || null;
}

Entity.prototype.hasRider = function (list) {
    if (this.isRideable()) for (let e of this.isRideable().getRiders()) {
        if (!e.isValid()) continue;
        if (typeof list === 'object') {
            for (let l of list) {
                if (isNaN(l)) {
                    if (e.typeId.includes(l)) return true;
                } else if (e.id === l) return true;
            }
        } else if (isNaN(list) && e.typeId.includes(list)) return true;
        else if (e.id == list) return true;
    }
    return false;
}

Entity.prototype.getFlyingSpeed = function () {
    return this.getComponent('minecraft:flying_speed').value;
}

Entity.prototype.setFlyingSpeed = function (s) {
    this.getComponent('minecraft:flying_speed').value = s;
}

Entity.prototype.color = function () {
    return this.getComponent('color') || null;
}

Entity.prototype.v3 = function () {
    return Vector.fromV3(this.location);
}

Entity.prototype.distanceTraveled = function () {
    if (!this._location || (this.getVelocity().x == 0 && this.getVelocity().y == 0 && this.getVelocity().z == 0))
        this._location = new Vector(this.location.x, this.location.y, this.location.z);
    return Vector.distance(this._location, this.location);
}

Entity.prototype.traveledDirection = function () {
    if (!this.t_location) this.t_location = new Vector(this.location.x, this.location.y, this.location.z);
    let d = Vector.calculateAngle(this.location, this.t_location);
    this.t_location = new Vector(this.location.x, this.location.y, this.location.z);
    return d;
}

Entity.prototype.addDataAsTag = function (name, value) {
    for (let t of this.getTags()) {
        if (t.startsWith(`${name}=`)) this.removeTag(t);
    }
    this.addTag(`${name}=${value}`);
}

Entity.prototype.getDataFromTag = function (name) {
    for (let t of this.getTags()) {
        if (!t.startsWith(`${name}=`)) continue;
        let val = t.split('=')[1];

        try {
            return JSON.parse(val);
        } catch (err) { };

        return val;
    }
    return;
}

Entity.prototype.removeDataFromTag = function (name) {
    for (let t of this.getTags()) {
        if (t.startsWith(`${name}=`)) this.removeTag(t);
    }
}

Entity.prototype.getAllDataNamesFromTag = function () {
    return this.getTags().filter(a => a.includes('='));
}

Entity.prototype.hasDataFromTag = function (name) {
    for (let t of this.getTags()) {
        if (t.startsWith(`${name}=`)) return true;
    }
    return false;
}

Entity.prototype.clearTag = function () {
    let tag = [];
    for (let t of this.getTags()) {
        tag.push(t);
        this.removeTag(t);
    }
    return tag;
}



Block.prototype.getState = function (state) {
    return this.permutation.getState(state);
}

Block.prototype.getAllState = function () {
    let states = {};
    for (let s in this.permutation.getAllStates()) {
        states[s] = this.permutation.getState(s);
    }
    return states;
}

Block.prototype.editState = function (state) {
    let states = {};
    for (let s in this.permutation.getAllStates()) {
        states[s] = this.permutation.getState(s);
    }
    for (let s in state) {
        states[s] = state[s];
    }
    return this.setPermutation(BlockPermutation.resolve(this.typeId, states));
}

Block.prototype.v3 = function () {
    return Vector.fromV3(this.location);
}





ItemStack.prototype.addLore = function (string) {
    return this.setLore(this.getLore().concat(string));
}
ItemStack.prototype.extract = function () {
    let data = {};
    data.typeId = this.typeId;
    data.amount = this.amount;
    data.lockMode = this.lockMode;
    data.nameTag = this.nameTag;
    data.lore = this.getLore();
    if (this.hasComponent('minecraft:durability')) data.damage = this.getComponent('minecraft:durability').damage;
    if (this.hasComponent('minecraft:enchantments')) {
        data.enchants = [];
        let en = this.getComponent('minecraft:enchantments').enchantments;
        let it = en[Symbol.iterator]();
        let result = it.next();
        while (!result.done) {
            data.enchants.push({
                type: result.value.type.id,
                level: result.value.level
            })
            result = it.next();
        }
    }
    return data;
}
ItemStack.prototype.setAmount = function (n) {
    let total = this.amount;
    if (typeof n === 'string' && !isNaN(n)) total += Number(n);
    else if (!isNaN(n)) total = n;
    if (total <= 0) return new ItemStack('minecraft:air');
    this.amount = total;
    return this;
}
ItemStack.compress = function (data) {
    let item = new ItemStack(data.typeId, data.amount);
    item.lockMode = data.lockMode;
    item.nameTag = data.nameTag;
    item.setLore(data.lore);
    if (item.hasComponent('minecraft:durability')) {
        if (data.damage != undefined) item.getComponent('minecraft:durability').damage = data.damage;
    }
    if (item.hasComponent('minecraft:enchantments')) {
        let en_list = item.getComponent('minecraft:enchantments').enchantments;
        for (let j of data.enchants) {
            en_list.addEnchantment(new Enchantment(j.type, j.level));
        }
        item.getComponent('minecraft:enchantments').enchantments = en_list;
    }
    return item;
}




Vector.fromV3 = function (v) {
    if (typeof v === 'object')
        return new Vector(v.x, v.y, v.z);
    return new Vector(0, 0, 0);
}

Vector.set = function (x, y, z) {
    if (typeof x === 'object') return new Vector(v.x, v.y, v.z);
    if (typeof x === 'number') return new Vector(x, y || x, z || x);
    return new Vector(0, 0, 0);
}

Vector.calculate = function (a, b, options = {}) {
    // Calculate the vector between the two points
    let vector = Vector.fromV3(a);
    if (typeof b === 'object') vector = Vector.fromV3(b).sub(a);

    // Calculate the distance between the points
    let distance = Math.sqrt(vector.x * vector.x + vector.y * vector.y + vector.z * vector.z);

    let yaw = Math.atan2(vector.z, vector.x);
    let pitch = Math.atan2(-vector.y, Math.sqrt(vector.x * vector.x + vector.z * vector.z));
    if (options.angle === 'degrees') {
        yaw *= (180 / Math.PI);
        //yaw = ((yaw + 90 + 180) % 360 + 360) % 360 - 180;
        pitch *= (180 / Math.PI);
    }


    return { yaw, pitch, distance };
}

Vector.prototype.clone = function () {
    return new Vector(this.x, this.y, this.z);
}

Vector.prototype.toString = function () {
    return `x:${this.x}, y:${this.y}, z:${this.z}`;
}

Vector.prototype.add = function (x, y, z) {
    if (typeof x === 'object') {
        this.x += x.x;
        this.y += x.y;
        this.z += x.z;
    } else {
        this.x += x;
        this.y += y || x;
        this.z += z || x;
    }
    return this;
}
Vector.add = function (v, x, y, z) {
    let t = v.clone();
    if (typeof x === 'object') {
        t.x += x.x;
        t.y += x.y;
        t.z += x.z;
    } else {
        t.x += x;
        t.y += y || x;
        t.z += z || x;
    }
    return t;
}

Vector.prototype.sub = function (x, y, z) {
    if (typeof x === 'object') {
        this.x -= x.x;
        this.y -= x.y;
        this.z -= x.z;
    } else {
        this.x -= x;
        this.y -= y || x;
        this.z -= z || x;
    }
    return this;
}
Vector.sub = function (v, x, y, z) {
    let t = v.clone();
    if (typeof x === 'object') {
        t.x -= x.x;
        t.y -= x.y;
        t.z -= x.z;
    } else {
        t.x -= x;
        t.y -= y || x;
        t.z -= z || x;
    }
    return t;
}

Vector.prototype.divide = function (x, y, z) {
    if (typeof x === 'object') {
        this.x /= x.x;
        this.y /= x.y;
        this.z /= x.z;
    } else {
        this.x /= x;
        this.y /= y || x;
        this.z /= z || x;
    }
    return this;
}
Vector.divide = function (v, x, y, z) {
    let t = v.clone();
    if (typeof x === 'object') {
        t.x /= x.x;
        t.y /= x.y;
        t.z /= x.z;
    } else {
        t.x /= x;
        t.y /= y || x;
        t.z /= z || x;
    }
    return t;
}

Vector.prototype.multiply = function (x, y, z) {
    if (typeof x === 'object') {
        this.x *= x.x;
        this.y *= x.y;
        this.z *= x.z;
    } else {
        this.x *= x;
        this.y *= y || x;
        this.z *= z || x;
    }
    return this;
}
Vector.multiply = function (v, x, y, z) {
    let t = v.clone();
    if (typeof x === 'object') {
        t.x *= x.x;
        t.y *= x.y;
        t.z *= x.z;
    } else {
        t.x *= x;
        t.y *= y || x;
        t.z *= z || x;
    }
    return t;
}

Vector.prototype.floor = function () {
    return new Vector(Math.floor(this.x), Math.floor(this.y), Math.floor(this.z));
}

Vector.prototype.getSides = function (filter = []) {
    let sides = [];
    if (!filter.includes('UP')) sides.push(this.clone().add(0, 1, 0)); // up;
    if (!filter.includes('DOWN')) sides.push(this.clone().add(0, -1, 0)); // down;
    if (!filter.includes('NORTH')) sides.push(this.clone().add(0, 0, -1)); // north;
    if (!filter.includes('SOUTH')) sides.push(this.clone().add(0, 0, 1)); // south;
    if (!filter.includes('EAST')) sides.push(this.clone().add(1, 0, 0)); // east;
    if (!filter.includes('WEST')) sides.push(this.clone().add(-1, 0, 0)); // west;
    return sides;
}

Vector.prototype.getMag = function () {
    return Math.pow(this.x * this.x + this.y * this.y + this.z * this.z, 0.5);
}

Vector.prototype.getDirection = function (target) {
    let v = target.sub(this.clone());
    let mag = v.getMag();
    return v.divide(mag);
}




export const OVERWORLD = world.getDimension('overworld');
export const NETHER = world.getDimension('nether');
export const THE_END = world.getDimension('the_end');



export class GUI {
    static actionForm(player, title, label, buttons = []) {
        const form = new ActionFormData();
        form.title(title || 'Custom GUI');
        let body = '\n';

        for (let l in label) {
            let val = label[l];
            if (l === 'br') {
                if (isNaN(val)) body += `\n`;
                else body += `\n`.repeat(val);
                continue;
            }
            let tmp = `${l}: `;
            switch (typeof val) {
                case 'number': tmp += `${val}`.red(); break;
                case 'boolean': tmp += `${val}`.blue(); break;
                case 'string': tmp += `${val}`.green(); break;
                case 'undefined': tmp += val; break;
            }
            body += `${tmp}`;
        }

        if (buttons.length == 0) {
            form.button('Done');
        } else for (let b of buttons) {
            if (typeof b !== 'string') continue;
            form.button(b);
        }
        form.body(body + '\n');
        return form.show(player);
    }
}







export const enchants = [
    'aqua_affinity',
    'bane_of_arthropods',
    'binding',
    'blast_protection',
    'channeling',
    'depth_strider',
    'efficiency',
    'feather_falling',
    'fire_aspect',
    'fire_protection',
    'flame',
    'fortune',
    'frost_walker',
    'impaling',
    'infinity',
    'knockback',
    'looting',
    'loyalty',
    'luck_of_the_sea',
    'lure',
    'mending',
    'multishot',
    'piercing',
    'power',
    'projectile_protection',
    'protection',
    'punch',
    'quick_charge',
    'respiration',
    'reptide',
    'sharpness',
    'silktouch',
    'smite',
    'soul_speed',
    'swift_sneak',
    'thorns',
    'unbreaking',
    'vanishing'
];

export const colors = [
    'black',
    'blue',
    'brown',
    'cyan',
    'gray',
    'green',
    'light_blue',
    'lime',
    'magenta',
    'orange',
    'pink',
    'purple',
    'red',
    'silver',
    'white',
    'yellow'
];



export function generateId(length) {
    const COMBI = ALPHABET + ALPHABET.toLowerCase() + NUMBERS + SPECIAL;
    return ' '.repeat(length).replace(/ /g, _ => COMBI.charAt(Math.floor(Math.random() * COMBI.length)));
}

export function generateUUID() {
    const COMBI = ALPHABET + ALPHABET.toLowerCase() + NUMBERS;
    return "########-####-####-####-############".replace(/#/g, _ => COMBI.charAt(Math.floor(Math.random() * COMBI.length)));
}

export function removeEntity(id) {
    let e = world.getEntity(id);
    if (e != undefined && e.isValid()) {
        e.remove();
        return true;
    }
    return false;
}

export function clamp(n, min, max) {
    return Math.min(Math.max(n, min), max);
}

export function random_bm() {
    let u = 0, v = 0;
    while (u === 0) u = Math.random(); //Converting [0,1) to (0,1)
    while (v === 0) v = Math.random();
    let num = Math.sqrt(-2.0 * Math.log(u)) * Math.cos(2.0 * Math.PI * v);
    num = num / 10.0 + 0.5; // Translate to 0 -> 1
    if (num > 1 || num < 0) return randn_bm() // resample between 0 and 1
    return num
}


export default {
    PI_180,

    sleep,
    world, system, Entity,
    Block, BlockPermutation,
    Vector, ItemStack,
    enchants, colors,

    OVERWORLD, NETHER, THE_END,
    generateId, generateUUID,
    removeEntity,
    clamp,

    random_bm
};